# encoding: utf-8

import io
import json
import os
import globalVars
import traceback

import speech
import api
import queueHandler
import threading
import tones
from tones import beep
import datetime

JGT_oldSpeak = speech.speakText
JGT_data=""
JGT_autoLastSpoken=False
JGT_autoClipboard = False
JGT_globalPlugin=None
JGT_LastTranslateTime = datetime.datetime.now()
JGT_File_Cache = None

lastSpokenInGame = False
def JGT_speak(text, *args, **kwargs):
	global JGT_data, JGT_oldSpeak, JGT_autoLastSpoken, JGT_globalPlugin, lastSpokenInGame, JGT_LastTranslateTime
	lastText = JGT_data
	JGT_data = text
	shouldSpeak = True
	if JGT_autoLastSpoken:
		if getInGame():
			if not lastSpokenInGame:
				lastSpokenInGame = True # this helps us avoid auto translating the title when they first switch into the game.
			else:
				# if this is the same text as before, and coming through faster than a user could press a key combo to repeat it, then it is probably someone using ClipReader with JGT.
				##ian 2018: I found a better way to stop double translating the same text, by only letting one translation web request thread be alive at any one time.
				#elapsed = (datetime.datetime.now() - JGT_LastTranslateTime).total_seconds()
				#if not (text == lastText and elapsed < .05):
				#if not (text == lastText):
					(shouldTranslate, shouldSpeak) = getShouldTranslate(text)
					if shouldTranslate:
						JGT_LastTranslateTime = datetime.datetime.now()
						#beep(500, 50)
						JGT_globalPlugin.startTranslateThread(text)
						#threading.Thread(target=JGT_globalPlugin.translate, args=(text,)).start()
	if shouldSpeak:
		JGT_oldSpeak(text, *args, **kwargs)

def is_ascii(s):
	return all(ord(c) < 256 for c in s)

def queueMessage(text):
	global JGT_oldSpeak
	queueHandler.queueFunction(queueHandler.eventQueue, JGT_oldSpeak, text)

def getShouldTranslate(text): # returns a boolean indicating if it should translate, and another indicating if we should speak the untranslated text.
	# returns (shouldTranslate, shouldSpeak)
	if not text or not isinstance(text,basestring) or text.isspace():
		return (False, True)
	# code for getting the title came from globalCommands.py in the NVDA source code.
	obj=api.getForegroundObject()
	title=obj.name
	if title == None:
		title = "" # making it an empty string saves us headaches later, like when we check its length.
	validGame = getInGame()
	if validGame and title == text:
		return (False, False) # we do not speak or translate the title when in a game because it is sometimes the entire menu.
	if validGame and title == text + " " + _("copied to clipboard"):
		return (False, True) # we will speak the copying of the title, but not translate it.
	if validGame and not is_ascii(text) and not text.endswith(_("selected")) and not text.startswith(_("selected")):
		return (True, False)
	return (False, True)

def getInGame():
	# returns a boolean
	obj=api.getForegroundObject()
	wcn = obj.windowClassName
	appName = obj.appModule.appName
	valid = wcn == "hspwnd0" or wcn == "oniwnd0" or appName == "play" or appName == "brave of cloudia"
	valid = valid or (wcn == "SDL_app" and appName == "ruby")
	return valid

def ianDebug():
	queueMessage(str(getInGame()))
	#obj=api.getForegroundObject()
	#title=obj.name
	#if title == None:
	#	title = "" # making it an empty string saves us headaches later, like when we check its length.
	#queueMessage(title)

def loadFileCache(language):
	global JGT_File_Cache
	path = os.path.join(globalVars.appArgs.configPath, "addons\\JGT\\Cache\\" + language + ".JGTCache")
	if not os.path.exists(path):
		file = io.open(path, mode="w", encoding="utf-8")
		file.write(u"[]")
		file.close()
	file = io.open(path, mode="r", encoding="utf-8")
	list = json.load(file)
	file.close()
	JGT_File_Cache = {}
	length = len(list)
	for i in range(0, length):
		original = list[i][1].strip("\0") # trimming the \0 character fixes the issue where many Brave of Cloudia messages, including in the options menu, were not matching anything in the offline cache.
		google = list[i][2]
		JGT_File_Cache[original] = google
		#if google == "Clipboard":
		#	logToOutputFile(google + u" = " + original)
		#	for blah in original:
		#		logToOutputFile(unicode(ord(blah)))
		#	if original in JGT_File_Cache:
		#		logToOutputFile(u"found it")
	#logToOutputFile(u"finished loading " + str(length) + " entries.")

def getTitle():
	# code for getting the title came from globalCommands.py in the NVDA source code.
	obj=api.getForegroundObject()
	title=obj.name
	if title == None:
		title = ""
	wcn = obj.windowClassName
	appName = obj.appModule.appName
	if wcn == "hspwnd0":
		if not title.startswith("Shadow Rine") and not title.startswith("Bokurano Daibouken") and not title.startswith("DREAMY TRAIN"):
			title = "hspwnd0" # we set the title to hspwnd0 when in BK3 because the title is useless for identifying the app.
			# The English version of Bokurano Daibouken 3 has a good title, but the other versions do not.
	if wcn == "oniwnd0":
		title = "oniwnd0"
	if appName == "brave of cloudia":
		title = "brave of cloudia"
	if wcn == "SDL_app" and appName == "ruby":
		title = "cave of marvels"
	if len(title) > 50:
		title = title[:50] # make sure the title won't cause any problems when saving a translation to the DB.
	return title

def getFromFileCache(text, toLang):
	global JGT_File_Cache
	if JGT_File_Cache == None:
		loadFileCache(toLang)
	try:
		if u"\n" in text:
			return TranslateMultipleLines(text)
		return Translate(text)
	except ValueError:
		return None
def TranslateMultipleLines(text):
	text = text.replace(u"\r", u"")
	parts = text.split(u"\n")
	for i in range(0, len(parts)):
		parts[i] = Translate(parts[i])
	return u"\n".join(parts)

def logTraceback():
	for line in traceback.format_stack():
		logToOutputFile(line.strip())
def Translate(text):
	#logTraceback()
	#text = ChangeUnicodeNumbersAndOtherSymbolsToAscii(text) moved this into the startTranslateThread method.
	title = getTitle()
	title = title.lower()
	if title == u"brave of cloudia":
		# if text starts with Japanese for item, or equipment, followed by many spaces, then we split on commas followed by a space.
		# same for item and equipment followed by a full stop.
		# or if the text contains the word belongings followed by a colon, and equipment followed by a colon.
		# or if the text starts with resistance full stop.
		# or if the text starts with a comma, it is probably telling you all the characters around you.
		# or if it starts with trick, magic full stop
		# or if it contains a right parenthesy followed by a full stop it could be part of some position menu that really needs to be split.
		# or if it ends with text that means indirect attack.
		# or it ends with text that means attack
		# or if it starts with Target wide colon
		# or it contains a right arrow surrounded by spaces, which often happens between a character name and all their equipment.
		# or it contains a full stop with at least 3 commas, which happens when listing skills or spells.
		if text.startswith(u"アイテム     ") or text.startswith(u"装備品     ") \
			or text.startswith(u"アイテム。") or text.startswith(u"装備品。") \
			or u"持ち物：" in text or u"装備：" in text \
			or text.startswith(u"耐性。") \
			or text.startswith(u",") \
			or text.startswith(u"技,魔法。") \
			or u")。" in text \
			or text.endswith(u"に間接攻撃した！") \
			or text.endswith(u"を攻撃！") \
			or text.startswith(u"対象：") \
			or u" → " in text \
			or HasFullStopFollowedByAtLeast3Commas(text):
			return SplitAndTranslateFullStops(text)
		# if text starts with party wide colon, we split on spaces, too bad it didn't have commas between like all the other places.
		if text.startswith(u"パーティ："):
			return SplitOnSpacesOnly(text)
	# end if title is "brave of cloudia"
	# check for was removed from the warehouse or  was put in a warehouse text found in BK3.
	if u"倉庫から取り出しました。" in text or u",倉庫に入れました。" in text:
		# this splits on full stops, commas, and colons, though we only need full stops and commas for this particular text.
		return SplitAndTranslateFullStops(text)
	# check for 0 hit range of the enemy colon, pretty sure it is from BK3. There are several relevant records around ID 54033.
	if u"0 命中範囲の敵 ：" in text:
		return SplitOnSpacesOnly(text)
	# check for text that contains the Japanese words block and number, not sure what game this is from, but check ID 62422 and the 30 or so records after it to see examples.
	if u"番ブロックに," in text and u"ダメージ。" in text:
		# we only need to split on full stops and commas for this text, exclamation marks might have been nice as well.
		return SplitAndTranslateFullStops(text)
	# we split on the Japanese word for "and" if the text ends with ", disappeared without permission came out without permission from the item bag."
	if text.endswith(u",アイテム袋から勝手に出てきて勝手に消えた。"):
		# check ID 76335 for an example, I expect it came from a BK game.
		return SplitOnTheWordAndOnly(text)
	return FinalTranslate(text)

def ChangeUnicodeNumbersAndOtherSymbolsToAscii(text):
	text = text.replace(unichr(0xff10), u"0")
	text = text.replace(unichr(0xff11), u"1")
	text = text.replace(unichr(0xff12), u"2")
	text = text.replace(unichr(0xff13), u"3")
	text = text.replace(unichr(0xff14), u"4")
	text = text.replace(unichr(0xff15), u"5")
	text = text.replace(unichr(0xff16), u"6")
	text = text.replace(unichr(0xff17), u"7")
	text = text.replace(unichr(0xff18), u"8")
	text = text.replace(unichr(0xff19), u"9")
	# now do comma, helpful in brave of cloudia
	text = text.replace(unichr(0x3001), u",")
	text = text.replace(unichr(0x30fb), u",") # turn a katakana middle dot into a comma.
	# and ideographic space
	text = text.replace(unichr(0x3000), u" ")
	return text

def HasFullStopFollowedByAtLeast3Commas(text):
	index = text.find(u"。")
	if index == -1:
		return False
	count = 0
	while True:
		if index + 1 >= len(text):
			break # don't index passed the end of the string.
		index = text.find(',', index + 1)
		if index == -1:
			break
		count += 1
	return count >= 3
def SplitAndTranslateFullStops(text):
	parts = text.split(u"。")
	for i in range(0, len(parts)):
		parts[i] = SplitAndTranslateWideColons(parts[i])
	return u". ".join(parts)
def SplitAndTranslateWideColons(text):
	parts = text.split(u"：")
	for i in range(0, len(parts)):
		parts[i] = SplitAndTranslateCommas(parts[i])
	return u": ".join(parts)
def SplitAndTranslateCommas(text):
	parts = text.split(u",")
	for i in range(0, len(parts)):
		parts[i] = SplitAndTranslateRightArrows(parts[i])
	return u", ".join(parts)
def SplitAndTranslateRightArrows(text):
	parts = text.split(u"→")
	for i in range(0, len(parts)):
		parts[i] = FinalTranslate(parts[i])
	return u" → ".join(parts)
def SplitOnSpacesOnly(text):
	parts = text.split(u" ")
	for i in range(0, len(parts)):
		parts[i] = FinalTranslate(parts[i])
	return u" ".join(parts)
def SplitOnTheWordAndOnly(text):
	parts = text.split(u"と")
	for i in range(0, len(parts)):
		parts[i] = FinalTranslate(parts[i])
	return u" and ".join(parts)

def FinalTranslate(text):
	global JGT_File_Cache
	# handle special cases
	if not HasSomeJapanese(text):
		# they didn't send us Japanese to begin with, so just send it back to them.
		return text

	text = text.strip()
	(text, originalNumbers) = ReplaceNumbers(text)
	#string combinedNamesWithSpaces = CombineNamesWithSpaces(ref text);
	output = None
	if text in JGT_File_Cache:
		logToOutputFile(u"original text " + text)
		output = JGT_File_Cache[text]
		logToOutputFile(u"translation " + output)
	else:
		logToOutputFile(u"Could not find " + text)
		raise ValueError("part of the string was not found in the cache")
	#if combinedNamesWithSpaces != null:
	#	UndoCombineNamesWithSpaces(ref output, combinedNamesWithSpaces)
	output = UndoNumberReplacement(output, originalNumbers)
	logToOutputFile(u"after UndoNumberReplacements " + output)
	return output

def ReplaceNumbers(input):
	# I've changed this to replace English letters in addition to numbers, to help solve the problem of people choosing different player names in BK3.
	# and then added the hyphen as well, since some people chose a name with an English hyphen in the middle.
	origNumbers = []
	output = []
	inNumber = False
	startIndex = -1
	nextNumberIndex = 0
	for i in range(0, len(input)):
		c = input[i]
		# had to add backslash because someone decided to put a backslash at the end of their name in one of the BK games.
		if IsDigit(c) or IsLetter(c) or c == u"-" or c == u"_" or c == u"\\":
			if not inNumber:
				# we just entered the start of the number.
				startIndex = i
				output.append(unicode(str(nextNumberIndex), 'utf-8'))
				nextNumberIndex += 1
				inNumber = True
			# otherwise we just skip this character
		else: # not a number
			output.append(c)
			if inNumber:
				# we were in a number, so we need to copy it out for the original numbers list.
				origNumbers.append(input[startIndex:i])
				inNumber = False
	# end for loop
	# special case: get the last number when the input text ends with a number
	if inNumber:
		origNumbers.append(input[startIndex:])
	return (u"".join(output), origNumbers)
def UndoNumberReplacement(text, originalNumbers):
	if len(originalNumbers) == 0:
		return text # no numbers need to be replaced
	output = text
	for i in range(0, len(originalNumbers)):
		#logToOutputFile(originalNumbers[i])
		output = output.replace(unicode(str(i), 'utf-8'), originalNumbers[i])
	# go in reverse order, so the number 10 gets replaced before the numbers 0 or 1.
	# for some reason it doesn't work right on the Brave of Cloudia load screen, oh well, we will just iterate forward instead.
	#for i in range(len(originalNumbers) - 1, -1, -1):
	#	output = output.replace(unicode(str(i), 'utf-8'), originalNumbers[i])
	return output

def IsDigit(c):
	# ascii digits 0 through 9 map to integer values 48 through 57 respectively
	c = ord(c)
	return 48 <= c and c <= 57
def IsLetter(c):
	# upper case letters map to integer values 65 through 90
	# lower case letters map to integer values 97 through 122
	c = ord(c)
	return (65 <= c and c <= 90) or (97 <= c and c <= 122)

def HasSomeJapanese(text):
	for i in range(0, len(text)):
		c = ord(text[i])
		if (0x3000 <= c and c <= 0x303f) \
			or (0x3040 <= c and c <= 0x309f) \
			or (0x30a0 <= c and c <= 0x30ff) \
			or (0xff00 <= c and c <= 0xffef) \
			or (0x4e00 <= c and c <= 0x9faf):
			return True
	return False
def logToOutputFile(text):
	return # disable logging
	path = os.path.join(globalVars.appArgs.configPath, "addons\\JGT\\Cache\\output.txt")
	file = io.open(path, mode="a", encoding="utf-8")
	file.write(text)
	file.write(u"\n")
	file.close()
